#include <stdexcept>
#include <set>
#include <algorithm>

#include "seq_span.hpp"
#include "mussa_exceptions.hpp"

SeqSpan::SeqSpan(const SeqSpan &o)
  : seq(o.seq),
    seq_start(o.seq_start),
    seq_count(o.seq_count),
    parent(o.parent),
    rc_seq(o.rc_seq)
{
}

SeqSpan::SeqSpan(const SeqSpan *p)
  : seq(p->seq),
    seq_start(p->seq_start),
    seq_count(p->seq_count),
    parent(p->parent),
    rc_seq(p->rc_seq) 
{
}

SeqSpan::SeqSpan(const std::string &seq_,
                 AlphabetRef a,
                 strand_type strand_)
  : seq_start(0),
    seq_count(seq_.length()),
    parent()
{
  switch (strand_) {
  case PlusStrand:
  case SingleStrand:
    // do nothing
    seq_strand = strand_;
    seq.reset(new SeqString(seq_, a));
  break;
  default:
    throw sequence_invalid_strand(
            "Can only initialize a Plus or Single strand from a string"
          );
    break;
  }
}

SeqSpan::SeqSpan(const SeqSpanRef parent_, 
                 size_type start_, 
                 size_type count_,
                 strand_type strand_) 
  : seq(parent_->seq),
    seq_start(parent_->seq_start + start_),
    parent(parent_)    
{
  if (count_ == npos)
    seq_count = parent_->seq_count;
  else
    seq_count = count_;
  
  // blech this strand logic is too long
  switch(strand_) {
  case UnknownStrand:
  case PlusStrand:
  case MinusStrand:
  case BothStrand:
    seq_strand = strand_;
  break;
  case SameStrand:
    if (parent) {
      seq_strand = parent->strand();
    } else {
      throw sequence_invalid_strand("SameStrand is undefined with no parent");
    }
  break;
  case OppositeStrand:
    if (parent) {
      switch (parent->strand()) {
      case PlusStrand:
        seq_strand = MinusStrand;
      break;
      case MinusStrand:
        seq_strand = PlusStrand;
      break;
      case UnknownStrand:
      case BothStrand:
        seq_strand = parent->strand();
      break;
      case SingleStrand:
        throw sequence_invalid_strand(
                "OppositeStrand is undefined with SingleStrands"
              );
        break;
      default:
        throw sequence_invalid_strand("Parent has an invalid strand type");
      }
    } else {
      throw sequence_invalid_strand("SameStrand is undefined with no parent");
    }
  break;
  case SingleStrand:
    if (parent) {
      if (parent->strand() == SingleStrand) {
        seq_strand = SingleStrand;
      } else {
        throw  sequence_invalid_strand("Can't change single strandedness");
      }
    } else {
      seq_strand = SingleStrand;
    }      
  break;
  default:
    throw sequence_invalid_strand("unrecognized strand identifier");
    break;
  }
}

//////
// Helpful operators
SeqSpan &SeqSpan::operator=(const SeqSpan& s)
{
  if (this != &s) {
    seq = s.seq;
    seq_start = s.seq_start;
    seq_count = s.seq_count;
    parent = s.parent;
    rc_seq = s.rc_seq;
  }
  return *this;
}

std::ostream& operator<<(std::ostream& out, const SeqSpan& s)
{
  out << s.sequence();
}

/* Not implemented yet
//! compare two spans
//! \throws sequence_invalid_comparison
friend bool operator<(const SeqSpan&, const SeqSpan&);
bool operator<(const SeqSpan& a, const SeqSpan& b)
{
  // are we subcomponents of the same sequence region?
  if (a.seq.get() == b.seq.get()) {
    if (a.seq_start < b.seq_start)
      return true;
    else
      return false;
  } else {
    throw mussa_error("can only compare two spans from the same sequence");
  }
}
*/

bool operator==(const SeqSpan& a, const SeqSpan& b)
{
  if (SeqSpan::isFamily(a, b)) {
    // can do fast comparison
    if (a.seq_start == b.seq_start and a.seq_count == b.seq_count) {
      // unknown strands match anything
      if (a.seq_strand == SeqSpan::UnknownStrand or 
          b.seq_strand == SeqSpan::UnknownStrand) {
        return true;
      } else { 
        // otherwise our stranded-ness needs to match
        return (a.seq_strand == b.seq_strand);
      }
    } else {
      return false;
    }
  }
  // if we're not part of the same seqspan heirarchy we're never equal
  // the string equality is handled off in Sequence
  return false;
}

bool operator!=(const SeqSpan& a, const SeqSpan& b)
{
  return not (a == b);
}

SeqSpan::const_reference SeqSpan::operator[](SeqSpan::size_type i) const
{
  return at(i);
}

SeqSpan::const_reference SeqSpan::at(SeqSpan::size_type i) const
{
  if (!seq) throw std::out_of_range("empty sequence");
  if (seq_strand == PlusStrand) {
    return seq->at(seq_start+i);
  } else {
    return seq->rc_map[seq->at((seq_start+seq_count-1)-i)];
  }
}

const char *SeqSpan::data() const
{
  if (seq) {
    return seq->c_str()+seq_start;
  } else 
    return 0;
}

SeqSpan::const_iterator SeqSpan::begin() const
{
  if (seq and seq_count != 0) {
    if (seq_strand != MinusStrand) {
      return seq->begin()+seq_start;
    } else {
      if (not rc_seq) {
        initialize_rc_seq();
      }
      return rc_seq->begin();
    }
  } 
  return SeqSpan::const_iterator(0);
}

SeqSpan::const_iterator SeqSpan::end() const
{
  if (seq and seq_count != 0) {
    if (seq_strand != MinusStrand) {
      return seq->begin() + seq_start + seq_count;
    } else {
      if (not rc_seq) {
        initialize_rc_seq();
      }
      return rc_seq->end();
    }
  } 
  return SeqSpan::const_iterator(0);
}

SeqSpan::const_reverse_iterator SeqSpan::rbegin() const
{
  if (seq and seq_count != 0) {
    if (seq_strand != MinusStrand) {
      return seq->rbegin()+(seq->size()-(seq_start+seq_count));
    } else {
      if (not rc_seq) {
        initialize_rc_seq();
      }
      return rc_seq->rbegin();
    }
  } 
  return SeqSpan::const_reverse_iterator();
}

SeqSpan::const_reverse_iterator SeqSpan::rend() const
{
  if (seq and seq_count != 0) {
    if (seq_strand != MinusStrand) {
      return rbegin() + seq_count;
    } else {
      if (not rc_seq) {
        initialize_rc_seq();
      }
      return rc_seq->rend();
    }
  }
  return SeqSpan::const_reverse_iterator();
}

bool SeqSpan::empty() const
{
  return (seq_count == 0) ? true : false;
}

SeqSpan::size_type SeqSpan::find_first_not_of(
  const std::string& query, 
  SeqSpan::size_type index) const
{
  typedef std::set<std::string::value_type> sequence_set;
  sequence_set match_set;
  
  for(std::string::const_iterator query_item = query.begin();
      query_item != query.end();
      ++query_item)
  {
    match_set.insert(*query_item);
  }  
  for(const_iterator base = begin();
      base != end();
      ++base)
  {
    if(match_set.find(*base) == match_set.end()) {
      return base-begin();
    } 
  }
  return SeqSpan::npos;
}
 
void SeqSpan::setStart(SeqSpan::size_type v)
{
  if (v > stop()) {
    // cry
    throw mussa_error("can't set Start > Stop");
  }
  seq_count += seq_start - v;
  seq_start = v;
}

void SeqSpan::setStop(SeqSpan::size_type v)
{
  if ( v < start() ) {
    // negative sized sequences are bad
    throw mussa_error("can't set Stop < Start");
  } 
  seq_count = std::min<size_type>(v - seq_start, parentSize()-parentStart());
}

SeqSpan::size_type SeqSpan::parentStart() const
{
  if (!parent) {
    // no parent
    return start();
  } else {
    return start() - parent->start();
  } 
}

void SeqSpan::setParentStart(SeqSpan::size_type v)
{
  setStart(parent->start() + v);
}

SeqSpan::size_type SeqSpan::parentStop() const
{
  if (!parent) {
    // no parent
    return stop();
  } else {
    return stop() - parent->start();
  }
}

void SeqSpan::setParentStop(SeqSpan::size_type v)
{
  setStop(parent->start() + v);
}

SeqSpanRef SeqSpan::subseq(size_type start, size_type count, strand_type strand)
{
  count = std::min<size_type>(count, seq_count - start);
    
  SeqSpanRef new_span(new SeqSpan(this->shared_from_this(), start, count, strand));
  return new_span;
}

std::string SeqSpan::sequence() const
{
  if (seq) {
    if (seq_strand != MinusStrand) 
      return seq->substr(seq_start, seq_count);
    else {
      initialize_rc_seq();
      return *rc_seq;
    } 
  } else { 
    return std::string();
  }
}

bool SeqSpan::isFamily(const SeqSpan& a, const SeqSpan& b)
{
  return a.seq.get() == b.seq.get(); 
}

void SeqSpan::initialize_rc_seq() const
{
  std::string new_rc_seq;
  std::string rc_map(seq->get_alphabet().get_complement_map());
  
  new_rc_seq.reserve(size());
  size_type i = (seq_start + seq_count);
  while( i != seq_start ) {
    --i;
    new_rc_seq.append(1, rc_map[ seq->at(i) ]);
  }
  
  SeqStringRef new_rc_ref(new SeqString(new_rc_seq, seq->get_alphabet_ref()));
  const_cast<SeqStringRef *>(&rc_seq)->swap(new_rc_ref);
}
