#include <QApplication>
#include <QClipboard>
#include <QDir>
#include <QFileDialog>
#include <QMessageBox>
#include <QMouseEvent>
#include <QRubberBand>
#include <QRect>
#include <QString>
#include <iostream>
#include <set>

#include <math.h>

#include "qui/seqbrowser/SequenceBrowser.hpp"
#include "mussa_exceptions.hpp"

using namespace std;

SequenceBrowser::SequenceBrowser(QWidget *parent)
  : QGLWidget(parent),
    rubberBand(0),
    popupMenu(0),
    copySelectedSequenceAsFastaAction(0),
    copySelectedSequenceAsStringAction(0),
    editSequencePropertiesAction(0)
{ 
  popupMenu = new QMenu(this);
  copySelectedSequenceAsFastaAction = new QAction(tr("&Copy as Fasta"), this);
  copySelectedSequenceAsStringAction = new QAction(tr("&Copy Sequence"), this);
  editSequencePropertiesAction = new QAction(tr("Sequence &Properties"), this);
  
  connect(copySelectedSequenceAsFastaAction, SIGNAL(triggered()), 
          this, SLOT(copySelectedSequenceAsFasta()));
  popupMenu->addAction(copySelectedSequenceAsFastaAction);
  copySelectedSequenceAsStringAction->setShortcut(Qt::CTRL | Qt::Key_C);
  connect(copySelectedSequenceAsStringAction, SIGNAL(triggered()), 
          this, SLOT(copySelectedSequenceAsString()));
  popupMenu->addAction(copySelectedSequenceAsStringAction);
  // connect edit properties action
  connect(editSequencePropertiesAction, SIGNAL(triggered()), 
          this, SLOT(editSequenceProperties()));
}

SequenceBrowser::SequenceBrowser(const SequenceBrowser& sb, QWidget *parent)
  : QGLWidget(parent),
    GlSeqBrowser(sb),
    rubberBand(sb.rubberBand),
    popupMenu(sb.popupMenu),
    copySelectedSequenceAsFastaAction(sb.copySelectedSequenceAsFastaAction)
{
  resize(sb.width(), sb.height());
  setZoom(sb.zoom());
  paintGL();
}

QMenu *SequenceBrowser::getPopupMenu()
{
  return popupMenu;
}

QAction *SequenceBrowser::getCopySelectedSequenceAsFastaAction()
{
  return copySelectedSequenceAsFastaAction;
}

QAction *SequenceBrowser::getCopySelectedSequenceAsStringAction()
{
  return copySelectedSequenceAsStringAction;
}

QAction *SequenceBrowser::getEditSequencePropertiesAction()
{
  return editSequencePropertiesAction;
}

QSize SequenceBrowser::sizeHint() const
{
  return QSize(viewportPixelWidth(), viewportPixelHeight());
}

void SequenceBrowser::setViewportCenter(float x)
{
  const float epsilon = 1e-10;
  float center = GlSeqBrowser::viewportCenter();
  float difference = fabsf(x - center);
  float abs_x = fabsf(x);
  center = fabsf(center);

  // the difference < epsilon * val is one of the recommended tests
  // for float equality.
  // of course since we're looking for not equals, we need to toss a
  // not at the beginning
  if (not (difference < epsilon * abs_x or difference < epsilon * center))
  {
    GlSeqBrowser::setViewportCenter(x);
    emit viewportChanged();
    update();
  }
}

void SequenceBrowser::setZoom(double new_zoom)
{
  if (new_zoom != GlSeqBrowser::zoom()) {
    GlSeqBrowser::setZoom(new_zoom);
    emit viewportChanged();
    update();
  }
}

void SequenceBrowser::setClipPlane(int )
{
/*
  if (clipZ != (double) newZ){
    clipZ = (double) newZ;
    update();
  }
*/
}

void SequenceBrowser::copySelectedSequenceAsFasta()
{
  // get fasta data
  std::string buffer;
  size_t base_pairs_copied = copySelectedTracksAsFasta(buffer);

  // get reference to clipboard
  QClipboard *clipboard = QApplication::clipboard();
  clipboard->setText(buffer.c_str());
  emit basepairsCopied(base_pairs_copied);
}

void SequenceBrowser::copySelectedSequenceAsString()
{
  // get fasta data
  std::string buffer;
  size_t base_pairs_copied = copySelectedTracksAsString(buffer);

  // get reference to clipboard
  QClipboard *clipboard = QApplication::clipboard();
  clipboard->setText(buffer.c_str());
  emit basepairsCopied(base_pairs_copied);
}

void SequenceBrowser::clear()
{
  GlSeqBrowser::clear();
  emit tracksChanged();
}

void SequenceBrowser::displayContextMenu(const QPoint& point)
{
  popupMenu->popup(point);
}

void SequenceBrowser::editSequenceProperties()
{
  // if there's a previous window, disconnect its signal
  if (properties) {
    disconnect(properties.get(), SIGNAL(propertiesChanged()),
               this, SLOT(updateGL()));
  }
  PropertiesWindowRef new_properties(new PropertiesWindow(track_container));
  properties = new_properties;
  connect(properties.get(), SIGNAL(propertiesChanged()),
          this, SLOT(updateGL()));
  properties->show();
}

void SequenceBrowser::push_sequence(boost::shared_ptr<Sequence> s)
{
  GlSeqBrowser::push_sequence(s);
  emit tracksChanged();
}

void SequenceBrowser::push_sequence(boost::shared_ptr<GlSequence> gs)
{
  GlSeqBrowser::push_sequence(gs);
  emit tracksChanged();
}

////////////////////
// Rendering code
void SequenceBrowser::initializeGL()
{
  GlSeqBrowser::initializeGL();
}

void SequenceBrowser::resizeGL(int width, int height)
{
  GlSeqBrowser::resizeGL(width, height);
  emit viewportChanged();
}

void SequenceBrowser::paintGL()
{
  GlSeqBrowser::paintGL();
}

void SequenceBrowser::mousePressEvent( QMouseEvent *e)
{
  switch(e->button()) {
    case Qt::LeftButton:
      startSelecting(e);
      break;
   case Qt::RightButton:
      break;
   default:
      break;
  }
}

void SequenceBrowser::mouseMoveEvent( QMouseEvent *e )
{
  if (rubberBand and rubberBand->isVisible()) {
    rubberBand->setGeometry(QRect(bandOrigin, e->pos()).normalized());
  }
}

void SequenceBrowser::mouseReleaseEvent( QMouseEvent *e)
{
  switch(e->button()) {
    case Qt::LeftButton:
      stopSelecting(e);
      break;
   case Qt::RightButton:
      // ok so selectedMode and drawing mode should probably be combinded
      // into a single state variable.
      if (rubberBand and 
          not rubberBand->isVisible() and
          selectedCanvasRegion.contains(e->pos())) {
        displayContextMenu(e->globalPos());
      }
      break;
   default:
      break;
  }

}

void SequenceBrowser::startSelecting(QMouseEvent *e)
{
  if (!rubberBand)
    rubberBand = new QRubberBand(QRubberBand::Rectangle, this);

  if (not rubberBand->isVisible()) {
    bandOrigin = e->pos();
    rubberBand->setGeometry(QRect(bandOrigin, QSize()));
    rubberBand->show();
  }
}

void SequenceBrowser::stopSelecting(QMouseEvent *e)
{
  if (rubberBand and rubberBand->isVisible()) {
    rubberBand->hide();
    selectedMode = false;
    QRect r = QRect(bandOrigin, e->pos()).normalized();
    bandOrigin = r.topLeft();
  
    selectRegion(r.top(), r.left(), r.bottom(), r.right());
    selectedCanvasRegion = r;
  }
}
